/*
 *
 *                 SUNIX Matrix Series Board Device Driver
 *
 *                               For Kernel 2.4.x
 *
 *
 *	Driver for SUNIX Matrix series board device driver
 *	Based on drivers/char/serial.c, by Linus Torvalds, Theodore Ts'o.
 *
 * 	This program is free software; you can redistribute it and/or modify
 * 	it under the terms of the GNU General Public License as published by
 * 	the Free Software Foundation; either version 2 of the License, or
 * 	(at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * 	GNU General Public License for more details.
 *
 * 	You should have received a copy of the GNU General Public License
 * 	along with this program; if not, write to the Free Software
 * 	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *                                                    		Version: 1.00
 *                                                          Date: 2007/01/31
 */
#ifdef 		MODVERSIONS
#ifndef 	MODULE
#define 	MODULE
#endif
#endif

#ifdef MODULE
#include <linux/config.h>
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#else
#define	MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif



#include <linux/version.h>
#include <linux/autoconf.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/serial_reg.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/smp_lock.h>
#include <linux/delay.h>


#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/wait.h>
#include <linux/tty_driver.h>
#include <linux/signal.h>
#include <linux/pci.h>
//#include <linux/workqueue.h>
#include <linux/circ_buf.h>



#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/segment.h>
#include <asm/serial.h>
#include <asm/bitops.h>

#include "sunix_matrix.h"


#ifdef CONFIG_PCI
#include <linux/pci.h>
#ifndef PCI_ANY_ID
#define PCI_ANY_ID (~0)
#endif
#endif 

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");


struct PORT
{
	int		bar;
	int		offset;
	int		length;
	int		intmask;
	int		flags;

};


struct PCI_BOARD
{
	unsigned int	vendorID;
	unsigned int 	deviceID;
	unsigned int	subVendorID;
	unsigned int	subDeviceID;
	unsigned int	numberOfPorts;
	unsigned char	ChipFlag;
	char			boardName[BOARDNAME_LENGTH];
	char			functionName[FUNCTIONNAME_LENGTH];

	struct PORT		port[MATRIX_UART_ONBOARD_MAX];

};


static struct PCI_BOARD matrix_pci_board_conf[]=
{

	// mode none
	{
		// VenID		DevID					SubVenID			SubSysID			Ports	ChipFlag				Name
		VENID_MATRIX,	DEVID_MATRIX_TYPE_I,	SUBVENID_MATRIX_PCI,	SUBDEVID_TEST,		1,		MATRIX_HW_ID_TYPE_I,	"P0000",  "None",
		{
			//	Bar		ofs		len		intmask		flags
			{	0,		0,		8,		0x01,		PF_S64	},	// UART 1
		},
	},


	// PCI interface
	// rs232
	// mode IPC-P1002
	{
		// VenID		DevID					SubVenID				SubSysID			Ports	ChipFlag				Name
		VENID_MATRIX,	DEVID_MATRIX_TYPE_I,	SUBVENID_MATRIX_PCI,	SUBDEVID_P1002,		2,		MATRIX_HW_ID_TYPE_I,	"P1002",  "RS-232",
		{
			//	Bar		ofs		len		intmask		flags
			{	0,		0,		8,		0x01,		PF_S64	},	// UART 1
			{	0,		8,		8,		0x02,		PF_S64	},	// UART 2
		},
	},


	// mode IPC-P1004
	{
		// VenID		DevID					SubVenID				SubSysID			Ports	ChipFlag				Name
		VENID_MATRIX,	DEVID_MATRIX_TYPE_I,	SUBVENID_MATRIX_PCI,	SUBDEVID_P1004,		4,		MATRIX_HW_ID_TYPE_I,	"P1004",  "RS-232",
		{
			//	Bar		ofs		len		intmask		flags
			{	0,		0,		8,		0x01,		PF_S64	},	// UART 1
			{	0,		8,		8,		0x02,		PF_S64	},	// UART 2
			{	0,		16,		8,		0x04,		PF_S64	},	// UART 3
			{	0,		24,		8,		0x08,		PF_S64	},	// UART 4
		},
	},


	// mode IPC-P1008
	{
		// VenID		DevID					SubVenID				SubSysID			Ports	ChipFlag				Name
		VENID_MATRIX,	DEVID_MATRIX_TYPE_I,	SUBVENID_MATRIX_PCI,	SUBDEVID_P1008,		8,		MATRIX_HW_ID_TYPE_I,	"P1008",  "RS-232",
		{
			//	Bar		ofs		len		intmask		flags
			{	0,		0,		8,		0x01,		PF_S64	},	// UART 1
			{	0,		8,		8,		0x02,		PF_S64	},	// UART 2
			{	0,		16,		8,		0x04,		PF_S64	},	// UART 3
			{	0,		24,		8,		0x08,		PF_S64	},	// UART 4
			{	0,		32,		8,		0x10,		PF_S64	},	// UART 5
			{	0,		40,		8,		0x20,		PF_S64	},	// UART 6
			{	0,		48,		8,		0x40,		PF_S64	},	// UART 7
			{	0,		56,		8,		0x80,		PF_S64	},	// UART 8
		},
	},






	// rs422/485
	// mode IPC-P2002
	{
		// VenID		DevID					SubVenID				SubSysID			Ports	ChipFlag				Name
		VENID_MATRIX,	DEVID_MATRIX_TYPE_I,	SUBVENID_MATRIX_PCI,	SUBDEVID_P2002,		2,		MATRIX_HW_ID_TYPE_I,	"P2002",  "RS-422/485",
		{
			//	Bar		ofs		len		intmask		flags
			{	0,		0,		8,		0x01,		PF_S64	},	// UART 1
			{	0,		8,		8,		0x02,		PF_S64	},	// UART 2
		},
	},


	// mode IPC-P2004
	{
		// VenID		DevID					SubVenID				SubSysID			Ports	ChipFlag				Name
		VENID_MATRIX,	DEVID_MATRIX_TYPE_I,	SUBVENID_MATRIX_PCI,	SUBDEVID_P2004,		4,		MATRIX_HW_ID_TYPE_I,	"P2004",  "RS-422/485",
		{
			//	Bar		ofs		len		intmask		flags
			{	0,		0,		8,		0x01,		PF_S64	},	// UART 1
			{	0,		8,		8,		0x02,		PF_S64	},	// UART 2
			{	0,		16,		8,		0x04,		PF_S64	},	// UART 3
			{	0,		24,		8,		0x08,		PF_S64	},	// UART 4
		},
	},


	// mode IPC-P2008
	{
		// VenID		DevID					SubVenID				SubSysID			Ports	ChipFlag				Name
		VENID_MATRIX,	DEVID_MATRIX_TYPE_I,	SUBVENID_MATRIX_PCI,	SUBDEVID_P2008,		8,		MATRIX_HW_ID_TYPE_I,	"P2008",  "RS-422/485",
		{
			//	Bar		ofs		len		intmask		flags
			{	0,		0,		8,		0x01,		PF_S64	},	// UART 1
			{	0,		8,		8,		0x02,		PF_S64	},	// UART 2
			{	0,		16,		8,		0x04,		PF_S64	},	// UART 3
			{	0,		24,		8,		0x08,		PF_S64	},	// UART 4
			{	0,		32,		8,		0x10,		PF_S64	},	// UART 5
			{	0,		40,		8,		0x20,		PF_S64	},	// UART 6
			{	0,		48,		8,		0x40,		PF_S64	},	// UART 7
			{	0,		56,		8,		0x80,		PF_S64	},	// UART 8
		},
	},






	// rs232/422/485
	// mode IPC-P3002
	{
		// VenID		DevID					SubVenID				SubSysID			Ports	ChipFlag				Name
		VENID_MATRIX,	DEVID_MATRIX_TYPE_I,	SUBVENID_MATRIX_PCI,	SUBDEVID_P3002,		2,		MATRIX_HW_ID_TYPE_I,	"P3002",	"RS-232/422/485",
		{
			//	Bar		ofs		len		intmask		flags
			{	0,		0,		8,		0x01,		PF_S64	},	// UART 1
			{	0,		8,		8,		0x02,		PF_S64	},	// UART 2
		},
	},


	// mode IPC-P3004
	{
		// VenID		DevID					SubVenID				SubSysID			Ports	ChipFlag				Name
		VENID_MATRIX,	DEVID_MATRIX_TYPE_I,	SUBVENID_MATRIX_PCI,	SUBDEVID_P3004,		4,		MATRIX_HW_ID_TYPE_I,	"P3004",	"RS-232/422/485",
		{
			//	Bar		ofs		len		intmask		flags
			{	0,		0,		8,		0x01,		PF_S64	},	// UART 1
			{	0,		8,		8,		0x02,		PF_S64	},	// UART 2
			{	0,		16,		8,		0x04,		PF_S64	},	// UART 3
			{	0,		24,		8,		0x08,		PF_S64	},	// UART 4
		},
	},


	// mode IPC-P3008
	{
		// VenID		DevID					SubVenID				SubSysID			Ports	ChipFlag				Name
		VENID_MATRIX,	DEVID_MATRIX_TYPE_I,	SUBVENID_MATRIX_PCI,	SUBDEVID_P3008,		8,		MATRIX_HW_ID_TYPE_I,	"P3008",	"RS-232/422/485",
		{
			//	Bar		ofs		len		intmask		flags
			{	0,		0,		8,		0x01,		PF_S64	},	// UART 1
			{	0,		8,		8,		0x02,		PF_S64	},	// UART 2
			{	0,		16,		8,		0x04,		PF_S64	},	// UART 3
			{	0,		24,		8,		0x08,		PF_S64	},	// UART 4
			{	0,		32,		8,		0x10,		PF_S64	},	// UART 5
			{	0,		40,		8,		0x20,		PF_S64	},	// UART 6
			{	0,		48,		8,		0x40,		PF_S64	},	// UART 7
			{	0,		56,		8,		0x80,		PF_S64	},	// UART 8
		},
	},
};



enum
{
// rs232
	MATRIX_P1002 = 1,
	MATRIX_P1004,
	MATRIX_P1008,

// rs422/485
	MATRIX_P2002,
	MATRIX_P2004,
	MATRIX_P2008,

// rs232/422/485
	MATRIX_P3002,
	MATRIX_P3004,
	MATRIX_P3008,
};



typedef struct
{
	unsigned short vendor;
	unsigned short device;
	unsigned short subvendor;
	unsigned short subdevice;
	unsigned short driver_data;
} matrix_pciInfo;



static matrix_pciInfo matrix_pci_board_id[] =
{
// rs232
	{VENID_MATRIX, DEVID_MATRIX_TYPE_I, SUBVENID_MATRIX_PCI, SUBDEVID_P1002, MATRIX_P1002},
	{VENID_MATRIX, DEVID_MATRIX_TYPE_I, SUBVENID_MATRIX_PCI, SUBDEVID_P1004, MATRIX_P1004},
	{VENID_MATRIX, DEVID_MATRIX_TYPE_I, SUBVENID_MATRIX_PCI, SUBDEVID_P1008, MATRIX_P1008},

// rs422/485
	{VENID_MATRIX, DEVID_MATRIX_TYPE_I, SUBVENID_MATRIX_PCI, SUBDEVID_P2002, MATRIX_P2002},
	{VENID_MATRIX, DEVID_MATRIX_TYPE_I, SUBVENID_MATRIX_PCI, SUBDEVID_P2004, MATRIX_P2004},
	{VENID_MATRIX, DEVID_MATRIX_TYPE_I, SUBVENID_MATRIX_PCI, SUBDEVID_P2008, MATRIX_P2008},

// rs232/422/485
	{VENID_MATRIX, DEVID_MATRIX_TYPE_I, SUBVENID_MATRIX_PCI, SUBDEVID_P3002, MATRIX_P3002},
	{VENID_MATRIX, DEVID_MATRIX_TYPE_I, SUBVENID_MATRIX_PCI, SUBDEVID_P3004, MATRIX_P3004},
	{VENID_MATRIX, DEVID_MATRIX_TYPE_I, SUBVENID_MATRIX_PCI, SUBDEVID_P3008, MATRIX_P3008},
	{0}
};



struct PORT_STATUS
{
	unsigned char	ier;
	unsigned char	iir;
	unsigned char	fcr;
	unsigned char	lcr;
	unsigned char	mcr;
	unsigned char	lsr;
	unsigned char	msr;

	unsigned long 	rx_count;
	unsigned long 	tx_count;
	unsigned int	errorflag;

};



struct MATRIX_DRIVER_CONF
{
	unsigned int			board_count;
	unsigned int			port_count;
	unsigned int			flags;
	spinlock_t				lock;
};
  


struct MATRIX_BOARD
{
	int						board_enum;
	int						board_number;
	unsigned int			bus_number;
	unsigned int			dev_number;
	struct PCI_BOARD		pb_info;
	unsigned int			ports;
	unsigned int			port_index;
	unsigned int			irq;
	unsigned int			vector;
	unsigned int			vector_mask;
	unsigned int			iobase[MATRIX_UART_ONBOARD_MAX];
	unsigned char			ChipFlag[MATRIX_UART_ONBOARD_MAX];
	unsigned int			flags;
	struct 	pci_dev			*pdev;
};


struct MATRIX_UART_PORT
{
	int						board_enum;
	unsigned int			bus_number;
	unsigned int			dev_number;
	struct PCI_BOARD		pb_info;
	unsigned int			port;
	unsigned int			irq;
	unsigned int			vector;
	unsigned int			vector_mask;
	unsigned int			iobase;
	//unsigned char __iomem	*membase;
	unsigned char			iotype;
	unsigned char			ChipFlag;
	unsigned int			flags;

	int						type;
	unsigned int			baud_rate;
	unsigned int			baud_base;
	struct PORT_STATUS		pstatus;
	unsigned char			flow_control;
	int						rx_trigger;
	int						xmit_fifo_size;
	int						custom_divisor;
	struct tty_struct		*tty;

	int						read_status_mask;
	int						ignore_status_mask;
	unsigned char			x_char;
	unsigned char			*tx_buf;
	int						tx_buf_cont;
	int						tx_buf_head;
	int						tx_buf_tail;

	int						close_delay;
	unsigned short			closing_wait;
	unsigned long			event;
	int						count;
	int						blocked_open;
	long					session;
	long					pgrp;
 	int						timeout;
 	unsigned char			ldisc_stop_rx;

	struct tq_struct 		tqueue;
    
	struct termios			normal_termios;
	struct termios			callout_termios;
        
	wait_queue_head_t 		open_wait;
	wait_queue_head_t 		close_wait;
	wait_queue_head_t 		delta_msr_wait;
	struct async_icount		icount;
	spinlock_t				lock;
};


static struct tty_driver			matrix_tty_sdriver;
static struct tty_driver			matrix_tty_cdriver;
static int		                    matrix_refcount;

static struct tty_struct			*matrix_tty[MATRIX_UART_TOTAL_MAX];
static struct termios				*matrix_termios[MATRIX_UART_TOTAL_MAX];
static struct termios				*matrix_termios_locked[MATRIX_UART_TOTAL_MAX];
static struct MATRIX_BOARD			matrix_pci_board_table[MATRIX_PCI_BOARDS_MAX];
static struct MATRIX_UART_PORT		matrix_port_table[MATRIX_UART_TOTAL_MAX];
static struct MATRIX_DRIVER_CONF	matrix_drvconf;

static int  baudrate_setup(struct MATRIX_UART_PORT *, unsigned int);
static int  termios_setup(struct MATRIX_UART_PORT *, struct termios *);
static int  matrix_startup(struct MATRIX_UART_PORT *);
static int  matrix_block_til_ready(struct tty_struct *, struct file *, struct MATRIX_UART_PORT *);
static void matrix_shutdown(struct MATRIX_UART_PORT *);
static void matrix_wait_until_sent(struct tty_struct *, int);
static void matrix_flush_chars(struct tty_struct *);
static int  matrix_open(struct tty_struct *, struct file *);
static void matrix_close(struct tty_struct *, struct file *);
static void matrix_put_char(struct tty_struct *, unsigned char);
static int  matrix_write_room(struct tty_struct *);
static int  matrix_chars_in_buffer(struct tty_struct *);
static void matrix_flush_buffer(struct tty_struct *);
static void matrix_throttle(struct tty_struct *);
static void matrix_unthrottle(struct tty_struct *);
static void matrix_stop(struct tty_struct *);
static void matrix_start(struct tty_struct *);
static void matrix_hangup(struct tty_struct *);
static void matrix_rs_break(struct tty_struct *, int);
static void matrix_send_break(struct MATRIX_UART_PORT *, int);
static int  matrix_get_uart_info(struct MATRIX_UART_PORT *, struct serial_struct *);
static int  matrix_set_uart_info(struct MATRIX_UART_PORT *, struct serial_struct *);
static int  matrix_get_lsr_info(struct MATRIX_UART_PORT *, unsigned int *);
static void matrix_set_termios(struct tty_struct *, struct termios *);
static int  matrix_ioctl(struct tty_struct *, struct file *, unsigned int, unsigned long);
static void matrix_do_softint(void *);
static int  matrix_register_tty_driver(void);
static int  matrix_unregister_tty_driver(void);
static int  matrix_board_init(void);
static int  matrix_get_pci_board_conf(void);
static int  matrix_pci_board_probe(void);
static int  matrix_pci_board_release(void);
static int  matrix_uart_init(struct MATRIX_UART_PORT *);
static int  matrix_get_modem_info(struct MATRIX_UART_PORT *, unsigned int *);
static int  matrix_set_modem_info(struct MATRIX_UART_PORT *, unsigned int, unsigned int *);

static void matrix_interrupt(int, void *, struct pt_regs *);

static int  matrix_write(struct tty_struct *, int, const unsigned char *, int);

static inline void matrix_modem_status_intr(struct MATRIX_UART_PORT *, unsigned char);
static inline void matrix_data_transmit(struct MATRIX_UART_PORT *);
static inline void matrix_data_receive(struct MATRIX_UART_PORT *, int *, unsigned char);

static unsigned char READ_INTERRUPT_VECTOR(int);
static unsigned char READ_UART_RX(unsigned int);
//static unsigned char READ_UART_IER(unsigned int);
static unsigned char READ_UART_IIR(unsigned int);
static unsigned char READ_UART_LCR(unsigned int);
static unsigned char READ_UART_MCR(unsigned int);
static unsigned char READ_UART_LSR(unsigned int);
static unsigned char READ_UART_MSR(unsigned int);
static void WRITE_UART_TX(unsigned int, unsigned char);
static void WRITE_UART_IER(unsigned int, unsigned char);
static void WRITE_UART_FCR(unsigned int, unsigned char);
static void WRITE_UART_LCR(unsigned int, unsigned char);
static void WRITE_UART_MCR(unsigned int, unsigned char);
static void WRITE_UART_DLL(unsigned int, int);
static void WRITE_UART_DLM(unsigned int, int);
static void enable_THRE(struct MATRIX_UART_PORT *);
static void disable_THRE(struct MATRIX_UART_PORT *);



static unsigned char READ_INTERRUPT_VECTOR(int inter_vector)
{
	unsigned char data;
	if (inter_vector)
	{
		data = inb(inter_vector);
		return data;
	}
	return 0;
}


// RX register operation
static unsigned char READ_UART_RX(unsigned int iobase)
{
	unsigned char data;
	if (iobase)
	{
		data = inb(iobase + UART_RX);
		return data;
	}
	return 0;
}


// TX register operation
static void WRITE_UART_TX(unsigned int iobase, unsigned char data)
{
	if (iobase)
	{
		outb(data, iobase + UART_TX);
	}
}


// IER register operation
/*
static unsigned char READ_UART_IER(unsigned int iobase)
{
	if (iobase)
	{
		return inb(iobase + UART_IER);
	}
	return 0;
}
*/


static void WRITE_UART_IER(unsigned int iobase, unsigned char data)
{
	if (iobase)
	{
		outb(data, iobase + UART_IER);
	}
}


// IIR register operation
static unsigned char READ_UART_IIR(unsigned int iobase)
{
	unsigned char data;
	if (iobase)
	{
		data = inb(iobase + UART_IIR);
		return data;
	}
	return 0;
}


// FCR register operation
static void WRITE_UART_FCR(unsigned int iobase, unsigned char data)
{
	if (iobase)
	{
		outb(data, iobase + UART_FCR);
	}
}


// LCR register operation
static unsigned char READ_UART_LCR(unsigned int iobase)
{
	unsigned char data;
	if (iobase)
	{
		data = inb(iobase + UART_LCR);
		return data;
	}
	return 0;
}


static void WRITE_UART_LCR(unsigned int iobase, unsigned char data)
{
	if (iobase)
	{
		outb(data, iobase + UART_LCR);
	}
}


// MCR register operation
static unsigned char READ_UART_MCR(unsigned int iobase)
{
	unsigned char data;
	if (iobase)
	{
		data = inb(iobase + UART_MCR);
		return data;
	}
	return 0;
}


static void WRITE_UART_MCR(unsigned int iobase, unsigned char data)
{
	if (iobase)
	{
		outb(data, iobase + UART_MCR);
	}
}


// LSR register operation
static unsigned char READ_UART_LSR(unsigned int iobase)
{
	unsigned char data;
	if (iobase)
	{
		data = inb(iobase + UART_LSR);
		return data;
	}
	return 0;
}


// MSR register operation
static unsigned char READ_UART_MSR(unsigned int iobase)
{
	unsigned char data;
	if (iobase)
	{
		data = inb(iobase + UART_MSR);
		return data;
	}
	return 0;
}


// divisor latch LSB register operation
static void WRITE_UART_DLL(unsigned int iobase, int data)
{
	if (iobase)
	{
		outb(data, iobase + UART_DLL);
	}
}


// divisor latch MSB register operation
static void WRITE_UART_DLM(unsigned int iobase, int data)
{
	if (iobase)
	{
		outb(data, iobase + UART_DLM);
	}
}


static void enable_THRE(struct MATRIX_UART_PORT *up)
{
	if (up->iobase)
	{
		up->pstatus.ier |= UART_IER_THRI;
		WRITE_UART_IER(up->iobase, up->pstatus.ier);
	}
}


static void disable_THRE(struct MATRIX_UART_PORT *up)
{
	if (up->iobase)
	{
		up->pstatus.ier &= ~UART_IER_THRI;
		WRITE_UART_IER(up->iobase, up->pstatus.ier);
	}
}


static int baudrate_setup(struct MATRIX_UART_PORT *up, unsigned int new_speed)
{
	int dlldlm = 0;
	int status = STATUS_SUCCESS;


	if (!up->tty || !up->tty->termios)
	{
		return status;
	}

	if (!up->iobase)
	{
		return status;
	}

   	up->baud_rate = new_speed;

	if (new_speed)
	{
	    dlldlm = (up->baud_base * 16) / (new_speed * 16);

	  	if (dlldlm == 0)
		{
	    	dlldlm = 1;
		}
	}
	else
	{
	    dlldlm = 0;
	}

	
	up->timeout = ((up->xmit_fifo_size * HZ * 10 * dlldlm) / up->baud_base);

	if (dlldlm)
	{
		up->pstatus.mcr |= UART_MCR_DTR;
		WRITE_UART_MCR(up->iobase, up->pstatus.mcr);
	}
	else
	{
		up->pstatus.mcr &= ~UART_MCR_DTR;
		WRITE_UART_MCR(up->iobase, up->pstatus.mcr);
    	return status;
	}

	up->pstatus.lcr = READ_UART_LCR(up->iobase);


	WRITE_UART_LCR(up->iobase, (up->pstatus.lcr | UART_LCR_DLAB));
	WRITE_UART_DLL(up->iobase, (dlldlm & 0xff));
	WRITE_UART_DLM(up->iobase, (dlldlm >> 8));
	WRITE_UART_LCR(up->iobase, up->pstatus.lcr);
	return status;
}


static int termios_setup(struct MATRIX_UART_PORT *up, struct termios *old_termios)
{
	unsigned int c_cflag;
	unsigned int c_iflag;
	unsigned int baudrate = 0;
    int status = STATUS_SUCCESS;

	if ((!up->tty) || (!up->tty->termios) || (!up->iobase))
	{
		return status;
	}

	c_cflag = up->tty->termios->c_cflag;
	c_iflag = up->tty->termios->c_iflag;

	switch (c_cflag & (CBAUD | CBAUDEX))
	{
        case B921600	: baudrate = 921600;	break;
        case B460800	: baudrate = 460800;	break;
        case B230400	: baudrate = 230400;	break;
        case B115200	: baudrate = 115200;	break;
        case B57600		: baudrate = 57600;		break;
        case B38400		: baudrate = 38400;		break;
        case B19200		: baudrate = 19200;		break;
        case B9600		: baudrate = 9600;		break;
        case B4800		: baudrate = 4800;		break;
        case B2400		: baudrate = 2400;		break;
		case B1800		: baudrate = 1800;		break;
        case B1200		: baudrate = 1200;		break;
		case B600		: baudrate = 600;		break;
        case B300		: baudrate = 300;		break;
		case B200		: baudrate = 200;		break;
		case B150		: baudrate = 150;		break;
		case B134		: baudrate = 134;		break;
        case B110		: baudrate = 110;		break;
		case B75		: baudrate = 75;		break;
		case B50		: baudrate = 50;		break;
        default			: baudrate = 9600;		break;
	}

    baudrate_setup(up, baudrate);
 
	switch (c_cflag & CSIZE)
	{
		case CS5: 	up->pstatus.lcr |= 0x00; break;
		case CS6: 	up->pstatus.lcr |= 0x01; break;
		case CS7: 	up->pstatus.lcr |= 0x02; break;
		case CS8: 	up->pstatus.lcr |= 0x03; break;
		default:  	up->pstatus.lcr |= 0x03; break;
	}

	if (c_cflag & CSTOPB)
	{
		up->pstatus.lcr |= 0x04;
	}

	if (c_cflag & PARENB)
	{
		if (c_cflag & PARODD)
		{
			up->pstatus.lcr |= 0x08;
		}
		else
		{
			up->pstatus.lcr |= 0x18;
		}
	}

	if (c_cflag & CMSPAR)
	{
		up->pstatus.lcr |= UART_LCR_SPAR;
	}

	
	up->pstatus.fcr |= UART_FCR_ENABLE_FIFO;

	if (up->ChipFlag == MATRIX_HW_ID_TYPE_I)
	{
		switch (up->rx_trigger)
		{
			case MATRIX_TYPE_I_TRIGGER_LEVEL_1: 	up->pstatus.fcr |= 0x21; break;
			case MATRIX_TYPE_I_TRIGGER_LEVEL_16: 	up->pstatus.fcr |= 0x61; break;
			case MATRIX_TYPE_I_TRIGGER_LEVEL_32: 	up->pstatus.fcr |= 0xa1; break;
			case MATRIX_TYPE_I_TRIGGER_LEVEL_56: 	up->pstatus.fcr |= 0xe1; break;
			default:  								up->pstatus.fcr |= 0x61; break;
		}
	}
	else
	{
		up->pstatus.fcr |= 0x21;
	}	

	up->pstatus.ier &= ~UART_IER_MSI;
	up->pstatus.mcr &= ~UART_MCR_AFE;

	if (c_cflag & CRTSCTS)
	{
		up->flow_control |= SUNIX_FLOWCTL_HW_AUTO;
		up->pstatus.ier |= UART_IER_MSI;
 		up->pstatus.mcr |= UART_MCR_AFE;
	}
	else
	{
		up->flow_control &= ~SUNIX_FLOWCTL_HW_AUTO;
	}


	if (c_cflag & CLOCAL)
	{
		up->flags &= ~ASYNC_CHECK_CD;
	}
	else
	{
		up->flags |= ASYNC_CHECK_CD;
		up->pstatus.ier |= UART_IER_MSI;
	}

	if ((c_iflag & IXON) || (c_iflag & IXOFF))
	{
		up->flow_control |= SUNIX_FLOWCTL_SW;
	}
	else
	{
		up->flow_control &= ~SUNIX_FLOWCTL_SW;
	}

	up->read_status_mask = UART_LSR_DR | UART_LSR_OE | UART_LSR_THRE;

	if (c_iflag & INPCK)
	{
		up->read_status_mask |= (UART_LSR_FE | UART_LSR_PE);
	}

	if (c_iflag & (BRKINT | PARMRK))
	{
	   	up->read_status_mask |= UART_LSR_BI;
	}

	up->ignore_status_mask = 0;

	if (c_iflag & IGNPAR)
	{
	    up->ignore_status_mask |= (UART_LSR_PE | UART_LSR_FE);
	}

	if (c_iflag & IGNBRK)
	{
		up->ignore_status_mask |= UART_LSR_BI;

		if (c_iflag & IGNPAR)
		{
			up->ignore_status_mask |= UART_LSR_OE;
		}
	}

	if ((c_cflag & CREAD) == 0)
	{
		up->ignore_status_mask |= UART_LSR_DR;
	}

	WRITE_UART_MCR(up->iobase, up->pstatus.mcr);
	WRITE_UART_IER(up->iobase, up->pstatus.ier);
	WRITE_UART_FCR(up->iobase, up->pstatus.fcr);
	WRITE_UART_LCR(up->iobase, up->pstatus.lcr);
	return status;
}      	


static int matrix_startup(struct MATRIX_UART_PORT *up)
{
	unsigned long page;
	unsigned long flag;
	int status = STATUS_SUCCESS;

	page = __get_free_page(GFP_KERNEL);
	
	if (!page)
	{
		status = -ENOMEM;
		return status;
	}

	spin_lock_irqsave(&up->lock, flag);

		
	if (up->flags & ASYNC_INITIALIZED)
	{
		free_page(page);
		spin_unlock_irqrestore(&up->lock, flag);
	    return status;
	}

	if (!up->iobase || !up->type)
	{
	    if (up->tty)
		{
			set_bit(TTY_IO_ERROR, &up->tty->flags);
		}

	    free_page(page);
		spin_unlock_irqrestore(&up->lock, flag);
	    return status;
	}

        	
	if (up->tx_buf)
	{
	    free_page(page);
	}
	else
	{
	   	up->tx_buf = (unsigned char *)page;
	}
		
	WRITE_UART_FCR(up->iobase, (UART_FCR_ENABLE_FIFO | UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT));	
	up->pstatus.lsr = READ_UART_LSR(up->iobase);
    
	if (up->pstatus.lsr == 0xff)
	{
		spin_unlock_irqrestore(&up->lock, flag);
    	if (capable(CAP_SYS_ADMIN))
		{
			if (up->tty)
			{
				set_bit(TTY_IO_ERROR, &up->tty->flags);
			}

			return status;
	   	}
		else
		{
			status = -ENODEV;
			return status;
		}
	}	
 
	(void) READ_UART_RX(up->iobase);
	(void) READ_UART_LSR(up->iobase);
	(void) READ_UART_IIR(up->iobase);
	(void) READ_UART_MSR(up->iobase);
	
	up->pstatus.mcr = UART_MCR_DTR | UART_MCR_RTS;

	WRITE_UART_MCR(up->iobase, up->pstatus.mcr);

  	if (up->tty)
	{
 		clear_bit(TTY_IO_ERROR, &up->tty->flags);
	}
	    
  	up->tx_buf_cont = up->tx_buf_head = up->tx_buf_tail = 0;

	spin_unlock_irqrestore(&up->lock, flag);

  	termios_setup(up, 0);	

	up->pstatus.ier = (UART_IER_MSI | UART_IER_RLSI | UART_IER_RDI);
	WRITE_UART_IER(up->iobase, up->pstatus.ier);

	(void) READ_UART_RX(up->iobase);
	(void) READ_UART_LSR(up->iobase);
	(void) READ_UART_IIR(up->iobase);
	(void) READ_UART_MSR(up->iobase);

	up->flags |= ASYNC_INITIALIZED;

	return status;
}


static int matrix_block_til_ready(struct tty_struct *tty, struct file *filp, struct MATRIX_UART_PORT *up)
{
	DECLARE_WAITQUEUE(wait, current);
	int status = STATUS_SUCCESS;
	int do_clocal = 0;
	unsigned long flag;
    
	if ((!up) || (!up->iobase))
	{
		status = -ENODEV;
		return status;
	}

     	
	if ((filp->f_flags & O_NONBLOCK) || (tty->flags & (1 << TTY_IO_ERROR))) 
	{

	    if (up->flags & ASYNC_CALLOUT_ACTIVE)
        {
            status = -EBUSY;
			return status;
        }

   	    up->flags |= ASYNC_NORMAL_ACTIVE;

   	    return status;
	}


	if (up->flags & ASYNC_CALLOUT_ACTIVE)
	{
	    if (up->normal_termios.c_cflag & CLOCAL)
        {
  		    do_clocal = 1;
        }
	}
	else
	{
	    if (tty->termios->c_cflag & CLOCAL)
        {
			do_clocal = 1;
        }
	}


	status = STATUS_SUCCESS;
	add_wait_queue(&up->open_wait, &wait);
	
	spin_lock_irqsave(&up->lock, flag);
	
	if (!tty_hung_up_p(filp))
    {
		up->count--;
    }

	spin_unlock_irqrestore(&up->lock, flag);

	up->blocked_open++;
	
	while (1)
	{
        spin_lock_irqsave(&up->lock, flag);

	    if (!(up->flags & ASYNC_CALLOUT_ACTIVE))
        {
		    up->pstatus.mcr = READ_UART_MCR(up->iobase);
		    WRITE_UART_MCR(up->iobase, (up->pstatus.mcr | UART_MCR_DTR | UART_MCR_RTS));
        }

		spin_unlock_irqrestore(&up->lock, flag);

   	    set_current_state(TASK_INTERRUPTIBLE);

   	    if (tty_hung_up_p(filp) || !(up->flags & ASYNC_INITIALIZED))
		{
			if (up->flags & ASYNC_HUP_NOTIFY)
            {
		  	    status = -EAGAIN;
            }
			else
            {
		        status = -ERESTARTSYS;
            }

			break;
   	    }
        
		up->pstatus.msr = READ_UART_MSR(up->iobase);
        
	    if (!(up->flags & ASYNC_CALLOUT_ACTIVE) &&
            !(up->flags & ASYNC_CLOSING) &&
            (do_clocal || (up->pstatus.msr & UART_MSR_DCD)))
        {
            break;
        }
				
	    if (signal_pending(current))
		{
		    status = -ERESTARTSYS;
		    break;
	    }
				
	    schedule();
	}
        
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&up->open_wait, &wait);
	
 	if (!tty_hung_up_p(filp))
    {
 		up->count++;
    }

	up->blocked_open--;
	
	if (status != STATUS_SUCCESS)
	{
		return status;
	}

	up->flags |= ASYNC_NORMAL_ACTIVE;

	return status;			
}


static void matrix_shutdown(struct MATRIX_UART_PORT *up)
{
	unsigned long flag;
    
	if ((!up) || (!up->iobase))
	{
		return;
	}
    
	if (!(up->flags & ASYNC_INITIALIZED))
    {
		return;
    }
	
	spin_lock_irqsave(&up->lock, flag);

	wake_up_interruptible(&up->delta_msr_wait);
	
	if (up->tx_buf)
	{
		free_page((unsigned long)up->tx_buf);
		up->tx_buf = NULL;
	}
	
	up->pstatus.ier = 0x00;
	WRITE_UART_IER(up->iobase, up->pstatus.ier);

	if (!up->tty || (up->tty->termios->c_cflag & HUPCL))
	{
		up->pstatus.mcr &= ~(UART_MCR_DTR | UART_MCR_RTS);
	}
    
	WRITE_UART_MCR(up->iobase, up->pstatus.mcr);

	WRITE_UART_FCR(up->iobase, (UART_FCR_ENABLE_FIFO | UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT));

	(void) READ_UART_RX(up->iobase);

	if (up->tty)
	{
		set_bit(TTY_IO_ERROR, &up->tty->flags);
	}

	up->flags &= ~ASYNC_INITIALIZED;

	spin_unlock_irqrestore(&up->lock, flag);
}


static void matrix_wait_until_sent(struct tty_struct *tty, int timeout)
{
	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)tty->driver_data;
	unsigned long orig_jiffies;
	unsigned long char_time;
    
	if ((!up) || (!up->iobase))
	{
		return;
	}
	
	if (up->type == PORT_UNKNOWN)
	{
		return;
	}

	if (up->xmit_fifo_size == 0)
	{
		return;
	}
		
	orig_jiffies = jiffies;
	char_time = (up->timeout - HZ/50) / up->xmit_fifo_size;
	char_time = char_time / 5;

	if (char_time == 0)
	{
		char_time = 1;
	}

	if (timeout && (timeout < char_time))
	{
		char_time = timeout;
	}
	
	if ((!timeout) || (timeout > (2 * up->timeout)))
	{
		timeout = 2 * up->timeout;
	}

    
	while (!((up->pstatus.lsr = READ_UART_LSR(up->iobase)) & UART_LSR_TEMT))
	{
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(char_time);

		if (signal_pending(current))
		{
			break;
		}

		if (timeout && time_after(jiffies, orig_jiffies + timeout))
		{
			break;
		}

	}

	set_current_state(TASK_RUNNING);		
}


static void matrix_flush_chars(struct tty_struct *tty)
{
	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)tty->driver_data;

	if ((!up) || (!up->iobase))
	{
		return;
	}

	if ((up->tx_buf_cont <= 0) || (tty->stopped) || (!up->tx_buf) || (tty->hw_stopped))
	{
		return;
	}

	enable_THRE(up);
}


static int matrix_uart_init(struct MATRIX_UART_PORT *up)
{
/*
	WRITE_UART_IER(up->iobase, 0x00);	// clear interrupt enable

	WRITE_UART_LCR(up->iobase, 0x80);	// set dlab
	WRITE_UART_DLL(up->iobase, 0x00);	// clear baudrate
	WRITE_UART_DLM(up->iobase, 0x00);
	WRITE_UART_LCR(up->iobase, 0x00);	// clear dlab

	WRITE_UART_MCR(up->iobase, 0x00);	// clear mcr
	WRITE_UART_FCR(up->iobase, 0x03);	// clear fcr
	WRITE_UART_FCR(up->iobase, 0x00);	// clear fcr
*/

    up->pstatus.rx_count = 0;
    up->pstatus.tx_count = 0;
    up->pstatus.errorflag = 0x00;    
	return STATUS_SUCCESS;
}


static int matrix_open(struct tty_struct *tty, struct file *filp)
{
	struct MATRIX_UART_PORT *up;
	int status = STATUS_SUCCESS;
	int line;

 	line = MINOR(tty->device) - tty->driver.minor_start;
    
	if ((line < 0) || (line >= MATRIX_UART_TOTAL_MAX))
	{
		status = -ENODEV;
		return status;
	}

	up = matrix_port_table + line;	

	if ((!up) || (!up->iobase))
	{
		status = -ENODEV;
		return status;
	}

 	tty->driver_data = up;
  	up->tty = tty;

	status = matrix_uart_init(up);
   	if (status)
	{
   		return status;
	}	

   	status = matrix_startup(up);

   	if (status)
	{
   		return status;
	}

  	status = matrix_block_til_ready(tty, filp, up);

  	if (status)
	{
  		return status;
	}
		
	up->count++;
    
	MOD_INC_USE_COUNT;
	
	if ((up->count == 1) && (up->flags & ASYNC_SPLIT_TERMIOS))
	{
   	    if (tty->driver.subtype == SERIAL_TYPE_NORMAL)
        {
			*tty->termios = up->normal_termios;
        }
   	    else
        {
			*tty->termios = up->callout_termios;
        }
	   	
		termios_setup(up, 0);
	}	
	
	up->session = current->session;	
 	up->pgrp = current->pgrp;
  
	clear_bit(TTY_DONT_FLIP, &tty->flags);
    
 	tty->low_latency = 1;
  
	return status;
}


static void matrix_close(struct tty_struct *tty, struct file *filp)
{
	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)tty->driver_data;
 	unsigned long timeout;
	unsigned long flag;
	int	line;
    
 	line = MINOR(tty->device) - tty->driver.minor_start;
    
	if ((line < 0) || (line >= MATRIX_UART_TOTAL_MAX))
    {
		return;
    }
		
	if ((!up) || (!up->iobase))
    {
   	    return;
    }

	spin_lock_irqsave(&up->lock, flag);

	if (tty_hung_up_p(filp))
	{
		spin_unlock_irqrestore(&up->lock, flag);
 		MOD_DEC_USE_COUNT;
		return;
	}			

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,21))			
	if ((tty->count == 1) && (up->count != 1))
	{ 	
#else
	if ((atomic_read(&tty->count) == 1) && (up->count != 1))
	{
#endif
   	    up->count = 1;
	}

	if (--up->count < 0)
    {
   	    up->count = 0;
    }
  
	if (up->count)
	{
		spin_unlock_irqrestore(&up->lock, flag);
   	    MOD_DEC_USE_COUNT;
   	    return;
	}
		
	up->flags |= ASYNC_CLOSING;
	spin_unlock_irqrestore(&up->lock, flag);
		
	if (up->flags & ASYNC_NORMAL_ACTIVE)
    {
	    up->normal_termios = *tty->termios;
    }

	if (up->flags & ASYNC_CALLOUT_ACTIVE)
    {
	    up->callout_termios = *tty->termios;
    }

	tty->closing = 1;
	
	if (up->closing_wait != ASYNC_CLOSING_WAIT_NONE)
    {
		tty_wait_until_sent(tty, up->closing_wait);
    }
		
	up->pstatus.ier &= ~UART_IER_RLSI;	
	
	if (up->flags & ASYNC_INITIALIZED)
	{
		WRITE_UART_IER(up->iobase, up->pstatus.ier);
        
	    timeout = jiffies + HZ;
	
	    while (!(READ_UART_LSR(up->iobase) & UART_LSR_TEMT))
		{
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(5);
			
			if (time_after(jiffies, timeout))
            {
		 		break;
            }
	    }
	}	

	matrix_shutdown(up);
	
	if (tty->driver.flush_buffer)
    {
		tty->driver.flush_buffer(tty);
    }

	if (tty->ldisc.flush_buffer)
    {
		tty->ldisc.flush_buffer(tty);
    }
			
	tty->closing = 0;
	up->event = 0;
	up->tty = NULL;
	
	if (up->blocked_open)
	{     
	    if (up->close_delay)
		{
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(up->close_delay);
	    }
      
	    wake_up_interruptible(&up->open_wait);
	}

	up->flags &= ~(ASYNC_NORMAL_ACTIVE | ASYNC_CALLOUT_ACTIVE | ASYNC_CLOSING);
	
	wake_up_interruptible(&up->close_wait);
	MOD_DEC_USE_COUNT;
}	


static int matrix_write(struct tty_struct *tty, int from_user, const unsigned char *buf, int count)
{
	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)tty->driver_data;
	unsigned long flag;
	int c;
	int total = 0;
	int status = STATUS_SUCCESS;

	if ((!up) || (!up->iobase))
	{
		status = -EFAULT;
	   	return status;
	}
    
	if ((!tty) || (!up->tx_buf))
	{
	   	return status;
	}
  
	
	while (1)
	{
 		c = MIN(count, MIN(SERIAL_XMIT_SIZE - up->tx_buf_cont - 1, SERIAL_XMIT_SIZE - up->tx_buf_head));
		if (c <= 0)
        {
			break;
        }

	   	if (from_user)
		{
			if (copy_from_user((up->tx_buf + up->tx_buf_head), buf, c) == c)
			{
				total = -EFAULT;
		    	break;
			}
	    }
		else
		{
			memcpy(up->tx_buf + up->tx_buf_head, buf, c);
		}
						
		spin_lock_irqsave(&up->lock, flag);
	    up->tx_buf_head   = (up->tx_buf_head + c) & (SERIAL_XMIT_SIZE - 1);
	    up->tx_buf_cont  += c;
		spin_unlock_irqrestore(&up->lock,flag);

	    buf   += c;
	    count -= c;
	  	total += c;
	}

    
	if ((up->tx_buf_cont) && (!tty->stopped) && (!(up->pstatus.ier & UART_IER_THRI)))
	{
		spin_lock_irqsave(&up->lock, flag);
		enable_THRE(up);
		spin_unlock_irqrestore(&up->lock,flag);
	}

	return total;
}


static void matrix_put_char(struct tty_struct *tty, unsigned char ch)
{
	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)tty->driver_data;
	unsigned long flag;
	            
	if ((!up) || (!up->iobase))
	{
	   	return;
	}

	if (!tty || !up->tx_buf)
	{
	   	return;
	}

	if (up->tx_buf_cont >= (SERIAL_XMIT_SIZE - 1))
	{
	   	return;
	}

	spin_lock_irqsave(&up->lock, flag);

	up->tx_buf[up->tx_buf_head++] = ch;
	up->tx_buf_head &= SERIAL_XMIT_SIZE - 1;
	up->tx_buf_cont++;

	spin_unlock_irqrestore(&up->lock, flag);
	
	if (!tty->stopped && !(up->pstatus.ier & UART_IER_THRI))
	{
		spin_lock_irqsave(&up->lock, flag);
		enable_THRE(up);
		spin_unlock_irqrestore(&up->lock, flag);
	}	
}


static int matrix_write_room(struct tty_struct *tty)
{
	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)tty->driver_data;
	int ret;
	int status = STATUS_SUCCESS;

	if ((!up) || (!up->iobase))
	{
		status = -EFAULT;
	   	return status;
	}

	ret = SERIAL_XMIT_SIZE - up->tx_buf_cont - 1;

	if (ret < 0)
	{
		ret = 0;
	}
	return ret;
}


static int matrix_chars_in_buffer(struct tty_struct *tty)
{
	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)tty->driver_data;
	int status = STATUS_SUCCESS;

	if ((!up) || (!up->iobase))
	{
		status = -EFAULT;
	   	return status;
	}

	return up->tx_buf_cont;	
}


static void matrix_flush_buffer(struct tty_struct *tty)
{
	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)tty->driver_data;
	unsigned long flag;

	if ((!up) || (!up->iobase))
	{
	   	return;
	}

	spin_lock_irqsave(&up->lock, flag);
	up->tx_buf_cont = up->tx_buf_head = up->tx_buf_tail = 0;
	spin_unlock_irqrestore(&up->lock, flag);

	WRITE_UART_FCR(up->iobase, (up->pstatus.fcr | UART_FCR_ENABLE_FIFO | UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT));
	WRITE_UART_FCR(up->iobase, up->pstatus.fcr);

	wake_up_interruptible(&tty->write_wait);

	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup)
	{
		(tty->ldisc.write_wakeup)(tty);
	}
}


static void matrix_throttle(struct tty_struct *tty)
{
	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)tty->driver_data;

	if ((!up) || (!up->iobase))
	{
	   	return;
	}

    up->ldisc_stop_rx = 1;

	if (up->flow_control & SUNIX_FLOWCTL_SW)
	{
		up->x_char = STOP_CHAR(tty);
		WRITE_UART_TX(up->iobase, up->x_char);
		up->icount.tx++;
		enable_THRE(up);
	}

	if (up->flow_control & SUNIX_FLOWCTL_HW_AUTO)
	{
 		up->pstatus.mcr &= ~UART_MCR_AFE;
 		up->pstatus.mcr &= ~UART_MCR_RTS;
		WRITE_UART_MCR(up->iobase, up->pstatus.mcr);
	}
}


static void matrix_unthrottle(struct tty_struct *tty)
{
	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)tty->driver_data;

	if ((!up) || (!up->iobase))
	{
	   	return;
	}

    up->ldisc_stop_rx = 0;

    if (up->flow_control & SUNIX_FLOWCTL_SW)
	{
		up->x_char = START_CHAR(tty);
		WRITE_UART_TX(up->iobase, up->x_char);
		up->icount.tx++;
		enable_THRE(up);
	}

	if (up->flow_control & SUNIX_FLOWCTL_HW_AUTO)
	{
 		up->pstatus.mcr |= UART_MCR_AFE;
 		up->pstatus.mcr |= UART_MCR_RTS;
		WRITE_UART_MCR(up->iobase, up->pstatus.mcr);
	} 
}


static void matrix_stop(struct tty_struct *tty)
{
	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)tty->driver_data;
	unsigned long flag;

	if ((!up) || (!up->iobase))
	{
	   	return;
	}

	spin_lock_irqsave(&up->lock, flag);

	if (up->pstatus.ier & UART_IER_THRI)
	{
		disable_THRE(up);
	}

	spin_unlock_irqrestore(&up->lock, flag);
}


static void matrix_start(struct tty_struct *tty)
{
	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)tty->driver_data;
	unsigned long flag;

	if ((!up) || (!up->iobase))
	{
	   	return;
	}

	spin_lock_irqsave(&up->lock, flag);

	if (up->tx_buf_cont && up->tx_buf && !(up->pstatus.ier & UART_IER_THRI))
	{
		enable_THRE(up);
	}

	spin_unlock_irqrestore(&up->lock, flag);
}


static void matrix_hangup(struct tty_struct *tty)
{
	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)tty->driver_data;

	if ((!up) || (!up->iobase))
	{
	   	return;
	}

	matrix_flush_buffer(tty);
	matrix_shutdown(up);
	up->event = 0;
	up->count = 0;
	up->flags &= ~(ASYNC_NORMAL_ACTIVE | ASYNC_CALLOUT_ACTIVE);
	up->tty = NULL;
	wake_up_interruptible(&up->open_wait);
}


static void matrix_rs_break(struct tty_struct *tty, int break_state)
{
	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)tty->driver_data;
	unsigned long flag;

	if ((!up) || (!up->iobase))
	{
	   	return;
	}

	spin_lock_irqsave(&up->lock, flag);

	if (break_state == -1)
	{
		up->pstatus.lcr = READ_UART_LCR(up->iobase);
		WRITE_UART_LCR(up->iobase, (up->pstatus.lcr | UART_LCR_SBC));
	}
	else
	{
		up->pstatus.lcr = READ_UART_LCR(up->iobase);
		WRITE_UART_LCR(up->iobase, (up->pstatus.lcr & ~UART_LCR_SBC));
	}

	spin_unlock_irqrestore(&up->lock, flag);
}


static inline void matrix_modem_status_intr(struct MATRIX_UART_PORT *up, unsigned char msr)
{
	if ((!up) || (!up->iobase))
	{
	   	return;
	}

	if (msr & UART_MSR_TERI)
	{
	    up->icount.rng++;
	}

	if (msr & UART_MSR_DDSR)
	{
	    up->icount.dsr++;
	}

	if (msr & UART_MSR_DDCD)
	{
	    up->icount.dcd++;
	}

	if (msr & UART_MSR_DCTS)
	{
	    up->icount.cts++;
	}

	wake_up_interruptible(&up->delta_msr_wait);
	
	if ((up->flags & ASYNC_CHECK_CD) && (msr & UART_MSR_DDCD))
	{
	 	if (msr & UART_MSR_DCD)
        {
			wake_up_interruptible(&up->open_wait);
        }
		else if (!((up->flags & ASYNC_CALLOUT_ACTIVE) && (up->flags & ASYNC_CALLOUT_NOHUP)))
		{
	       	set_bit(SUNIX_EVENT_HANGUP, &up->event);
		}
			
		MOD_INC_USE_COUNT;
        
		if (schedule_task(&up->tqueue) == 0)
        {
	 		MOD_DEC_USE_COUNT;
        }
	}

	if (up->flow_control & SUNIX_FLOWCTL_HW_AUTO) 
	{
		if (up->tty->hw_stopped)
		{
			if (msr & UART_MSR_CTS)
			{
		   	    up->tty->hw_stopped = 0;

		   		if ((up->type != PORT_16750) && (!up->ChipFlag))
				{
					enable_THRE(up);
		   		}
               
	    	    set_bit(SUNIX_EVENT_TXLOW, &up->event);        

				MOD_INC_USE_COUNT;
                
				if (schedule_task(&up->tqueue) == 0)
                {
   	 			    MOD_DEC_USE_COUNT;
                }
     	    }
	    }
		else
		{
			// when auto flow control, don't care cts signal
			if (!(msr & UART_MSR_CTS))
			{
				/*
		    	up->tty->hw_stopped = 1;
		   		if ((up->type != PORT_16750) && (!up->ChipFlag)) 
				{
					disable_THRE(up);
		   		}
				*/
			}
	    }
	}
}


static inline void matrix_data_transmit(struct MATRIX_UART_PORT *up)
{
	struct tty_struct *tty;
	int count;
	int cnt;
	int tail_tmp = 0;
	unsigned char data;
	unsigned long flag;

	if ((!up) || (!up->iobase))
	{
	   	return;
	}

	tty = up->tty;
	spin_lock_irqsave(&up->lock, flag);

	if ((up->tx_buf == 0))
	{
		spin_unlock_irqrestore(&up->lock, flag);
	    return;
	}

	if ((up->tx_buf_cont <= 0) || (up->tty->stopped) || (tty->hw_stopped))
	{
		disable_THRE(up);
		spin_unlock_irqrestore(&up->lock, flag);
		return;
	}

	cnt = up->tx_buf_cont;
	count = up->xmit_fifo_size; 		
	
	do
	{
		tail_tmp = up->tx_buf_tail;
		data = up->tx_buf[up->tx_buf_tail++];
		WRITE_UART_TX(up->iobase, data);
	    up->tx_buf_tail = up->tx_buf_tail & (SERIAL_XMIT_SIZE - 1);
			
	    if ((--up->tx_buf_cont <= 0))
		{
			break;
		}

	} while (--count > 0);

	up->icount.tx += (cnt - up->tx_buf_cont);
		
 	if (up->tx_buf_cont < WAKEUP_CHARS)
	{
 		set_bit(SUNIX_EVENT_TXLOW, &up->event);
		MOD_INC_USE_COUNT;

  	    if (schedule_task(&up->tqueue) == 0)
        {
 			MOD_DEC_USE_COUNT;
        }
 	}	

	if (up->tx_buf_cont <= 0)
	{
		disable_THRE(up);
	}

	spin_unlock_irqrestore(&up->lock, flag);
}


static inline void matrix_data_receive(struct MATRIX_UART_PORT *up, int *status, unsigned char iir)
{
	struct tty_struct *tty = up->tty;
	int recv_room = 0;
	int max_count = 256;

	unsigned char lsr = *status;
	unsigned char data;
	unsigned long flag;
	char errflag;

	if ((!up) || (!up->iobase))
	{
	   	return;
	}
  	
	recv_room = tty->ldisc.receive_room(tty);

 	if ((recv_room == 0) && (!up->ldisc_stop_rx))
	{
 		matrix_throttle(tty);
		return;
 	}

	spin_lock_irqsave(&up->lock, flag);

	do
	{
		data = READ_UART_RX(up->iobase);
		errflag = TTY_NORMAL;
		up->icount.rx++;
		up->pstatus.rx_count++;

		if (up->flow_control & SUNIX_FLOWCTL_SW)
		{
			if (data == STOP_CHAR(up->tty))
			{
				up->tty->stopped = 1;
				enable_THRE(up);
				goto IGNORE_CHAR;
			}

			if (data == START_CHAR(up->tty))
			{
				up->tty->stopped = 0;
				enable_THRE(up);
				goto IGNORE_CHAR;
			}
		}

		if (unlikely(lsr & (UART_LSR_BI | UART_LSR_FE | UART_LSR_PE | UART_LSR_OE)))
		{
			if (lsr & UART_LSR_BI)
			{
				lsr &= ~(UART_LSR_FE | UART_LSR_PE);
				up->icount.brk++;

				if (up->flags & ASYNC_SAK)
				{
					do_SAK(up->tty);
				}
			}
			else if (lsr & UART_LSR_PE)
			{
				up->icount.parity++;
			}
			else if (lsr & UART_LSR_FE)
			{
				up->icount.frame++;
			}

			if (lsr & UART_LSR_OE)
			{
				up->icount.overrun++;
				//WRITE_UART_FCR(up->iobase, up->pstatus.fcr | 0x23);
			}


			lsr &= up->read_status_mask;

			if (lsr & UART_LSR_BI)
			{
				errflag = TTY_BREAK;
			}
			else if (lsr & UART_LSR_PE)
			{
				errflag = TTY_PARITY;
			}
			else if (lsr & UART_LSR_FE)
			{
				errflag = TTY_FRAME;
			}
		}


		if ((lsr & up->ignore_status_mask & ~UART_LSR_OE) == 0)
		{
			tty_insert_flip_char(tty, data, errflag);
		}

		if (lsr & ~up->ignore_status_mask & UART_LSR_OE)
		{
			tty_insert_flip_char(tty, 0, TTY_OVERRUN);
		}

IGNORE_CHAR:
		lsr = up->pstatus.lsr = READ_UART_LSR(up->iobase);

	} while ((lsr & UART_LSR_DR) && (max_count-- > 0));
    

	tty_flip_buffer_push(tty);

	spin_unlock_irqrestore(&up->lock, flag);

	*status = lsr;
}


static void matrix_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct MATRIX_UART_PORT *up;
	struct MATRIX_UART_PORT *first_up;

	int i;
	int max;
	int irqbits;
	int bits;
	int pass_counter = 0;
	int handled = 0;

	int lsr;
	unsigned char msr;
	unsigned char iir;

	first_up = NULL;

    for (i = 0; i < MATRIX_PCI_BOARDS_MAX; i++)
	{
        if (dev_id == &(matrix_port_table[i * MATRIX_UART_ONBOARD_MAX]))
		{
           	first_up = dev_id;
			break;
        }
	}
			
    if (i == MATRIX_PCI_BOARDS_MAX)
	{
       	goto INTR_END;
	}

    if (first_up == 0)
	{
       	goto INTR_END;
	}
	
    max = matrix_pci_board_table[i].pb_info.numberOfPorts;

	{
		while (1)
		{
			irqbits = READ_INTERRUPT_VECTOR(first_up->vector) & first_up->vector_mask;

	    	if (irqbits == 0x00)
			{
				break;
			}

	    	handled = 1;

	    	for (i = 0, bits = 1; i < max; i++, bits <<= 1)
			{

				if (!(bits & irqbits))
				{
					continue;
				}

				up = first_up + i;
				iir = up->pstatus.iir = READ_UART_IIR(up->iobase) & 0x0f;

				if (iir & UART_IIR_NO_INT)
				{
					continue;
				}

				lsr = up->pstatus.lsr = READ_UART_LSR(up->iobase);

				if (lsr & UART_LSR_OE)
				{
					up->pstatus.errorflag |= MATRIX_ERR_OE;
				}

				if (lsr & UART_LSR_BI)
				{
					up->pstatus.errorflag |= MATRIX_ERR_BI;
				}

				if (lsr & UART_LSR_FE)
				{
					up->pstatus.errorflag |= MATRIX_ERR_FE;
				}

				if (lsr & UART_LSR_PE)
				{
					up->pstatus.errorflag |= MATRIX_ERR_PE;
				}

				if ((iir == UART_IIR_RLSI) || (iir == UART_IIR_CTO) || (iir == UART_IIR_RDI))
				{
					matrix_data_receive(up, &lsr, iir);
				}

				if ((iir == UART_IIR_THRI) && (lsr & UART_LSR_THRE))
				{
					matrix_data_transmit(up);
				}

				msr = up->pstatus.msr = READ_UART_MSR(up->iobase);

				if (msr & UART_MSR_ANY_DELTA)
				{
 					matrix_modem_status_intr(up, msr);
				}
			}

	    	if (pass_counter++ > INTERRUPT_COUNT)
			{
				break;
			}
		}
	}

INTR_END:
	handled = 1;
}


static void matrix_send_break(struct MATRIX_UART_PORT *up, int duration)
{
	unsigned long flag;

	if ((!up) || (!up->iobase))
	{
		return;
	}

	set_current_state(TASK_INTERRUPTIBLE);

	spin_lock_irqsave(&up->lock, flag);

	up->pstatus.lcr = READ_UART_LCR(up->iobase);
	WRITE_UART_LCR(up->iobase, (up->pstatus.lcr | UART_LCR_SBC));

	schedule_timeout(duration);

	up->pstatus.lcr = READ_UART_LCR(up->iobase);
	WRITE_UART_LCR(up->iobase, (up->pstatus.lcr & ~UART_LCR_SBC));

	spin_unlock_irqrestore(&up->lock, flag);
}


static int matrix_get_modem_info(struct MATRIX_UART_PORT *up, unsigned int *value)
{
    unsigned int result;
    unsigned long flag;
    int status = STATUS_SUCCESS;



    spin_lock_irqsave(&up->lock, flag);

    up->pstatus.msr = READ_UART_MSR(up->iobase);
	
 	spin_unlock_irqrestore(&up->lock, flag);

	if (up->pstatus.msr & UART_MSR_ANY_DELTA)
    {
        matrix_modem_status_intr(up, up->pstatus.msr);
    }

 	result = ((up->pstatus.mcr & UART_MCR_RTS) ? TIOCM_RTS : 0) |
             ((up->pstatus.mcr & UART_MCR_DTR) ? TIOCM_DTR : 0) |
			 ((up->pstatus.msr & UART_MSR_DCD) ? TIOCM_CAR : 0) |
		 	 ((up->pstatus.msr & UART_MSR_RI)  ? TIOCM_RNG : 0) |
		 	 ((up->pstatus.msr & UART_MSR_DSR) ? TIOCM_DSR : 0) |
		 	 ((up->pstatus.msr & UART_MSR_CTS) ? TIOCM_CTS : 0);

	put_user(result, (unsigned long *)value);

	return status;
}


static int matrix_set_modem_info(struct MATRIX_UART_PORT *up, unsigned int cmd, unsigned int *value)
{
	unsigned int arg;  
	int status = STATUS_SUCCESS;

	status = verify_area(VERIFY_READ, value, sizeof(int));

	if (status)
    {
        return status;
    }

	get_user(arg, (unsigned long *)value);

	switch (cmd)
	{
		case TIOCMBIS:
        {
            if (arg & TIOCM_RTS)
            {
			    up->pstatus.mcr |= UART_MCR_RTS;
            }
          
	        if (arg & TIOCM_DTR)
            {
		        up->pstatus.mcr |= UART_MCR_DTR;
            }
	        break;
        }
            
		case TIOCMBIC:
        {
	        if (arg & TIOCM_RTS)
            {
		        up->pstatus.mcr &= ~UART_MCR_RTS;
            }
          
	        if (arg & TIOCM_DTR)
            {
			    up->pstatus.mcr &= ~UART_MCR_DTR;
            }
	        break;
        }

		case TIOCMSET:
        {
		    up->pstatus.mcr &= ~(UART_MCR_RTS|UART_MCR_DTR);
			up->pstatus.mcr = ((arg & TIOCM_RTS)?UART_MCR_RTS:0) | ((arg & TIOCM_DTR)?UART_MCR_DTR:0);
	    	break;
        }

		default:
        {
            status = -EINVAL;
	    	return status;
        }
	}

    WRITE_UART_MCR(up->iobase, up->pstatus.mcr);
	return status;
}


static int matrix_get_uart_info(struct MATRIX_UART_PORT *up, struct serial_struct *serial_return)
{
	struct serial_struct serial_local;
	int status = STATUS_SUCCESS;

	if ((!up) || (!up->iobase))
	{
		status = -EFAULT;
		return status;
	}

	if (!serial_return)
	{
		status = -EFAULT;
		return status;
	}

	memset(&serial_local, 0, sizeof(serial_local));

	serial_local.type 			= up->type;
	serial_local.line 			= up->port;
	serial_local.port 			= up->iobase;
	serial_local.irq  			= up->irq;
	serial_local.flags 			= up->flags;
	serial_local.baud_base 		= up->baud_base;
	serial_local.close_delay 	= up->close_delay;
	serial_local.closing_wait 	= up->closing_wait;
	serial_local.custom_divisor = up->custom_divisor;
	serial_local.hub6 			= 0;

	if (copy_to_user(serial_return, &serial_local, sizeof(*serial_return)))
	{
		status = -EFAULT;
		return status;
	}

	return status;
}


static int matrix_set_uart_info(struct MATRIX_UART_PORT *up, struct serial_struct *serial_new)
{
	struct serial_struct serial_local;
	unsigned int flags;
	int status = STATUS_SUCCESS;


	if ((!serial_new) || (!up) || (!up->iobase))
	{
		status = -EFAULT;
		return status;
	}

	if (copy_from_user(&serial_local, serial_new, sizeof(serial_local)))
	{
		status = -EFAULT;
		return status;
	}

	if ((serial_local.irq != up->irq) ||
		(serial_local.port != up->iobase) ||
		(serial_local.custom_divisor != up->custom_divisor) ||
		(serial_local.baud_base != up->baud_base))
	{
		status = -EPERM;
		return status;
	}

	flags = up->flags & ASYNC_SPD_MASK;

    if (!capable(CAP_SYS_ADMIN))
	{
		if ((serial_local.baud_base != up->baud_base) ||
			(serial_local.close_delay != up->close_delay) ||
			((serial_local.flags & ~ASYNC_USR_MASK) != (up->flags & ~ASYNC_USR_MASK)))
		{
			status = -EPERM;
			return status;
		}

	    up->flags = ((up->flags & ~ASYNC_USR_MASK) | (serial_local.flags & ASYNC_USR_MASK));
	}
	else
	{
	    up->flags = ((up->flags & ~ASYNC_FLAGS) | (serial_local.flags & ASYNC_FLAGS));
	    up->close_delay = serial_local.close_delay * HZ/100;
	    up->closing_wait = serial_local.closing_wait * HZ/100;
	    up->tty->low_latency = (up->flags & ASYNC_LOW_LATENCY) ? 1 : 0;
	    up->tty->low_latency = 0;
	    up->type = serial_local.type;
	}

	if (up->ChipFlag == MATRIX_HW_ID_TYPE_I)
	{
		up->xmit_fifo_size	= MATRIX_TYPE_I_FIFOSIZE_SET;
		up->rx_trigger		= MATRIX_TYPE_I_TRIGGER_LEVEL_SET;
	}
	else
	{
		up->xmit_fifo_size	= DEFAULT_FIFOSIZE;
		up->rx_trigger		= DEFAULT_TRIGGER_LEVEL;
	}	


	if (up->flags & ASYNC_INITIALIZED)
	{
	    if (flags != (up->flags & ASYNC_SPD_MASK))
		{
			termios_setup(up, 0);
		}
	}
	else
	{
	   	status = matrix_startup(up);
	}
		
	return status;
}


static int matrix_get_lsr_info(struct MATRIX_UART_PORT *up, unsigned int *value)
{
	unsigned char lsr;
	unsigned long flag;
	int status = STATUS_SUCCESS;

	if ((!up) || (!up->iobase))
	{
		status = -EFAULT;
		return status;
	}

	spin_lock_irqsave(&up->lock, flag);

	lsr = up->pstatus.lsr = READ_UART_LSR(up->iobase);

	spin_unlock_irqrestore(&up->lock, flag);


	if (copy_to_user(value, &lsr, sizeof(int)))
	{
		status = -EFAULT;
		return status;
	}

	return status;
}


static void matrix_set_termios(struct tty_struct *tty, struct termios *old_termios)
{
	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)tty->driver_data;

	if ((!up) || (!up->iobase))
	{
		return;
	}

	if ((tty->termios->c_cflag != old_termios->c_cflag)|| (RELEVANT_IFLAG(tty->termios->c_iflag) != RELEVANT_IFLAG(old_termios->c_iflag)))
	{
		termios_setup(up, old_termios);
	}

	if ((old_termios->c_cflag & CRTSCTS) && !(tty->termios->c_cflag & CRTSCTS))
	{
 	   	tty->hw_stopped = 0;
		matrix_start(tty);
	}

	if ((old_termios->c_iflag & IXON) && !(tty->termios->c_iflag & IXON))
	{
 		tty->stopped = 0;
     	matrix_start(tty);
	}
}


static int matrix_ioctl(struct tty_struct *tty, struct file *file, unsigned int cmd, unsigned long arg)
{
 	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)tty->driver_data;
 	struct async_icount cprev;
 	struct async_icount cnow;
 	struct serial_icounter_struct icount;
	struct port_info dump;

 	int status = STATUS_SUCCESS;
	int tmp;
	int retvalue;
	unsigned long flag;


	if ((!up) || (!up->iobase))
	{
		status = -EFAULT;
		return status;
	}

    
	if ((cmd != TIOCGSERIAL) && (cmd != TIOCMIWAIT) && (cmd != TIOCGICOUNT))
	{
		if (tty->flags & (1 << TTY_IO_ERROR))
		{
			status = -EIO;
			return status;
		}
	}
	
	switch (cmd)
	{
		case TCSBRK:
		{
	    	status = tty_check_change(tty);
	   		if (status)
			{
				return status;
			}

	    	tty_wait_until_sent(tty, 0);

			if (signal_pending(current))
			{
				status = -EINTR;
				return status;
			}

	    	if (!arg)
			{
				matrix_send_break(up, HZ/4);
				if (signal_pending(current))
				{
					status = -EINTR;
					return status;
				}
			}

			return status;
		}
        
		
		case TCSBRKP:
		{
	   		status = tty_check_change(tty);
	   		if (status)
			{
				return status;
			}

	   		tty_wait_until_sent(tty, 0);

			if (signal_pending(current))
			{
				status = -EINTR;
				return status;
			}

	    	matrix_send_break(up, arg ? arg*(HZ/10) : HZ/4);

			if (signal_pending(current))
			{
				status = -EINTR;
				return status;
			}

			return status;
		}

        
		case TIOCGSOFTCAR: 
		{
			tmp = C_CLOCAL(tty) ? 1 : 0;
			retvalue = copy_to_user((void *)arg, &tmp, sizeof(int));

			if (retvalue)
			{
				status = -EFAULT;
				return status;
			}

			return status;
		}

        
		case TIOCSSOFTCAR: 
		{
			if (copy_from_user(&tmp, (void *)arg, sizeof(int)))
			{
				status = -EFAULT;
				return status;
			}
			tty->termios->c_cflag =	((tty->termios->c_cflag & ~CLOCAL) | (tmp ? CLOCAL : 0));

			return status;
		}

        
		case TIOCMGET:
        {
	    	status = verify_area(VERIFY_WRITE, (void *)arg,	sizeof(unsigned int));

	    	if (status)
            {
					return status;
            }
            
	    	return(matrix_get_modem_info(up, (unsigned int *)arg));
        }

         
		case TIOCMBIS: 
		case TIOCMBIC:  
		case TIOCMSET:
    		return(matrix_set_modem_info(up, cmd, (unsigned int *)arg));

      
		case TIOCGSERIAL:
		{
			return matrix_get_uart_info(up, (struct serial_struct *)arg);
		}


		case TIOCSSERIAL:
		{
			return matrix_set_uart_info(up, (struct serial_struct *)arg);
		}


		case TIOCSERGETLSR:
		{
			return matrix_get_lsr_info(up, (unsigned int *)arg);
		}


		case TIOCMIWAIT:
		{
 	        save_flags(flag); 
		    cli();

		    cprev = up->icount;
		    restore_flags(flag);

		    up->pstatus.ier |= UART_IER_MSI;
            WRITE_UART_IER(up->iobase, up->pstatus.ier);

		    while (1)
			{
			    interruptible_sleep_on(&up->delta_msr_wait);
			    if (signal_pending(current))
                {
                    status = -ERESTARTSYS;
				    return status;
                }
                
			    save_flags(flag);
				cli();

				cnow = up->icount; 
				restore_flags(flag);

				if (cnow.rng == cprev.rng && cnow.dsr == cprev.dsr && cnow.dcd == cprev.dcd && cnow.cts == cprev.cts)
                {
                    status = -EIO;
				    return status;
                }

				if (((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
				   	((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
				   	((arg & TIOCM_CD)  && (cnow.dcd != cprev.dcd)) ||
				   	((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)))
				{
				    return status;
				}

                cprev = cnow;
			}
		}

        
		case TIOCGICOUNT: 
		{
			spin_lock_irqsave(&up->lock, flag);

			cnow = up->icount;

			spin_unlock_irqrestore(&up->lock, flag);

			icount.cts         = cnow.cts;
			icount.dsr         = cnow.dsr;
			icount.rng         = cnow.rng;
			icount.dcd         = cnow.dcd;
			icount.rx          = cnow.rx;
			icount.tx          = cnow.tx;
			icount.frame       = cnow.frame;
			icount.overrun     = cnow.overrun;
			icount.parity 	   = cnow.parity;
			icount.brk         = cnow.brk;
			icount.buf_overrun = cnow.buf_overrun;

			if (copy_to_user((void *)arg, &icount, sizeof(icount)))
			{
				status = -EFAULT;
				return status;
			}

			return status;
		}					

        
		case SUNIX_DUMP_PORT_INFO:  
		{
			int count;

			for (count = 0; count < BOARDNAME_LENGTH; count++)
			{
				dump.board_name_info[count] = up->pb_info.boardName[count];
			}

			for (count = 0; count < FUNCTIONNAME_LENGTH; count++)
			{
				dump.function_name_info[count] = up->pb_info.functionName[count];
			}

			dump.bus_number_info = up->bus_number;
			dump.dev_number_info = up->dev_number;
			dump.port_info	     = up->port;
			dump.base_info	     = up->iobase;
			dump.irq_info 	     = up->irq;

			if (copy_to_user((void *)arg, &dump, sizeof(dump)))
			{
				status = -EFAULT;
				return status;
			}

			return status;
		}	
								
		default:
		{
			status = -ENOIOCTLCMD;
			return status;
		}
	}

	return status;
}


static void matrix_do_softint(void *private_)
{

	struct MATRIX_UART_PORT *up = (struct MATRIX_UART_PORT *)private_;
	struct tty_struct *tty;

	if ((!up) || (!up->iobase))
	{
		return;
	}

	tty = up->tty;

	if (tty)
	{
		if (test_and_clear_bit(SUNIX_EVENT_TXLOW, &up->event))
		{
	    	if ((tty->flags &  (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup)
			{
				(tty->ldisc.write_wakeup)(tty);
			}
	        wake_up_interruptible(&tty->write_wait);
	    }

	    if (test_and_clear_bit(SUNIX_EVENT_HANGUP, &up->event))
		{
			tty_hangup(tty);
		}
	}
    
    MOD_DEC_USE_COUNT;
}


static int matrix_register_tty_driver(void)
{
	int status = STATUS_SUCCESS;

	matrix_drvconf.flags &= ~DRIVER_TTY_REGISTER_ERR;

	memset(&matrix_tty_sdriver, 0, sizeof(struct tty_driver));


	matrix_tty_sdriver.magic 				= TTY_DRIVER_MAGIC;
	matrix_tty_sdriver.name 				= "ttySMX";
	matrix_tty_sdriver.major 				= SUNIX_MATRIX_TTY_MAJOR,
	matrix_tty_sdriver.minor_start 			= 0;
	matrix_tty_sdriver.num 					= MATRIX_UART_TOTAL_MAX;
	matrix_tty_sdriver.type 				= TTY_DRIVER_TYPE_SERIAL,
	matrix_tty_sdriver.subtype 				= SERIAL_TYPE_NORMAL,
	matrix_tty_sdriver.flags 				= TTY_DRIVER_REAL_RAW,
	matrix_tty_sdriver.init_termios 		= tty_std_termios;
	matrix_tty_sdriver.init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;
  	matrix_tty_sdriver.init_termios.c_iflag = 0;

    matrix_tty_sdriver.refcount             = &matrix_refcount;
	matrix_tty_sdriver.table				= matrix_tty;
    
	matrix_tty_sdriver.termios				= matrix_termios;
	matrix_tty_sdriver.termios_locked		= matrix_termios_locked;

	matrix_tty_sdriver.open					= matrix_open;
	matrix_tty_sdriver.close				= matrix_close;
	matrix_tty_sdriver.write				= matrix_write;
	matrix_tty_sdriver.put_char				= matrix_put_char;
	matrix_tty_sdriver.flush_chars			= matrix_flush_chars;
	matrix_tty_sdriver.write_room			= matrix_write_room;
	matrix_tty_sdriver.chars_in_buffer		= matrix_chars_in_buffer;
	matrix_tty_sdriver.flush_buffer			= matrix_flush_buffer;
	matrix_tty_sdriver.ioctl				= matrix_ioctl;
	matrix_tty_sdriver.throttle				= matrix_throttle;
	matrix_tty_sdriver.unthrottle			= matrix_unthrottle;
	matrix_tty_sdriver.set_termios			= matrix_set_termios;
	matrix_tty_sdriver.stop					= matrix_stop;
	matrix_tty_sdriver.start				= matrix_start;
	matrix_tty_sdriver.hangup				= matrix_hangup;
 
	matrix_tty_sdriver.break_ctl 			= matrix_rs_break;
	matrix_tty_sdriver.wait_until_sent 		= matrix_wait_until_sent;

	matrix_tty_cdriver						= matrix_tty_sdriver;
	matrix_tty_cdriver.name					= "cmx";
	matrix_tty_cdriver.major				= SUNIX_MATRIX_CAL_MAJOR;
	matrix_tty_cdriver.subtype				= SERIAL_TYPE_CALLOUT;
    
	status = tty_register_driver(&matrix_tty_sdriver);

	if (status != STATUS_SUCCESS)
	{
		matrix_drvconf.flags |= DRIVER_TTY_REGISTER_ERR;
		printk("SUNIX Error: tty driver register fail !\n");
		return status;
	}

	status = tty_register_driver(&matrix_tty_cdriver);

	if (status != STATUS_SUCCESS)
	{
		matrix_drvconf.flags |= DRIVER_TTY_REGISTER_ERR;
		printk("SUNIX Error: tty driver register fail !\n");
		return status;
	}
	return status;
}


static int matrix_unregister_tty_driver(void)
{
	int status = STATUS_SUCCESS;

	status = tty_unregister_driver(&matrix_tty_sdriver);

	if (status != STATUS_SUCCESS)
	{
		printk("SUNIX Error: tty driver unregister fail !\n");
	}

	status = tty_unregister_driver(&matrix_tty_cdriver);

	if (status != STATUS_SUCCESS)
	{
		printk("SUNIX Error: tty driver unregister fail !\n");
	}
    
	return status;
}


static int matrix_board_init(void)
{
	struct MATRIX_BOARD *mb = NULL;
	struct MATRIX_UART_PORT *up = NULL;
    int status = STATUS_SUCCESS;
	int i;
	int j;
	int n;

	memset(matrix_port_table, 0, MATRIX_UART_TOTAL_MAX * sizeof(struct MATRIX_UART_PORT));

	for (i = 0; i < MATRIX_PCI_BOARDS_MAX; i++)
	{
		mb = &matrix_pci_board_table[i];

		if (mb->board_enum > 0)
		{
			n = mb->port_index = (mb->board_number * MATRIX_UART_ONBOARD_MAX);
			up = &matrix_port_table[n];

			if (up == NULL)
			{
				status = STATUS_PORT_TBL_ADDRESS_ERR;
				printk("SUNIX Error: SUNIX port table address error !\n");
				return status;
			}

			for (j = 0; j < mb->ports; j++, n++, up++)
			{
				up->board_enum		= mb->board_enum;
				up->bus_number		= mb->bus_number;
				up->dev_number		= mb->dev_number;
				up->baud_base		= 921600;
				up->pb_info			= mb->pb_info;
				up->type			= PORT_16750;
				up->flow_control    = SUNIX_FLOWCTL_NONE;
				up->count			= 0;
				up->port 			= n;
				up->iobase			= mb->iobase[j];
				up->irq 			= mb->irq;
				up->vector 			= mb->vector;
				up->vector_mask 	= mb->vector_mask;
				up->ldisc_stop_rx	= 0;
				up->ChipFlag 		= mb->ChipFlag[j];
				up->flags 			= ASYNC_SHARE_IRQ;

				up->custom_divisor 	= 14745600;
				up->close_delay 	= 5 * HZ / 10;
				up->closing_wait 	= 30 * HZ;
               
		        up->normal_termios	= matrix_tty_sdriver.init_termios;
		        up->callout_termios	= matrix_tty_cdriver.init_termios;
        
		        up->tqueue.routine	= matrix_do_softint;
		        up->tqueue.data		= up;
      
				init_waitqueue_head(&up->open_wait);
				init_waitqueue_head(&up->close_wait);
				init_waitqueue_head(&up->delta_msr_wait);
				spin_lock_init(&up->lock);

				if (up->ChipFlag == MATRIX_HW_ID_TYPE_I)
				{
					up->xmit_fifo_size 	= MATRIX_TYPE_I_FIFOSIZE_SET;
					up->rx_trigger		= MATRIX_TYPE_I_TRIGGER_LEVEL_SET;
					up->iotype			= MATRIX_TYPE_IO;
				}
				else
				{
					up->xmit_fifo_size	= DEFAULT_FIFOSIZE;
					up->rx_trigger		= DEFAULT_TRIGGER_LEVEL;
					up->iotype			= MATRIX_TYPE_IO;
				}

			}

			n = mb->port_index = (mb->board_number * MATRIX_UART_ONBOARD_MAX);
			up = &matrix_port_table[n];

			if (up == NULL)
			{
				status = STATUS_PORT_TBL_ADDRESS_ERR;
				printk("SUNIX Error: SUNIX port table address error !\n");
				return status;
			}

			status = request_irq(mb->irq, matrix_interrupt, SA_SHIRQ, "sunix_matrix", up);

			if (status)
			{
				matrix_drvconf.flags |= DRIVER_PCI_IRQ_REGISTER_ERR;
	 			printk("SUNIX Error: SUNIX Matrix Board (bus:%d device:%d) request irq %d fail, IRQ %d may be conflit with another device.\n",
								mb->bus_number,
								mb->dev_number,
								mb->irq,
								mb->irq
								);
				return status;
			}

			matrix_drvconf.port_count = matrix_drvconf.port_count + mb->ports;
		}
	}
    return status;
}


static int matrix_get_pci_board_conf(void)
{
	struct MATRIX_BOARD *mb = NULL;
	struct pci_dev *pdev = NULL;
	int status = STATUS_SUCCESS;
	int i;
	int j;

	for (i = 0; i < MATRIX_PCI_BOARDS_MAX; i++)
	{
		mb = &matrix_pci_board_table[i];

		if (mb->board_enum > 0)
		{
			pdev = mb->pdev;
			mb->ports = mb->pb_info.numberOfPorts;
			mb->iobase[0] = pci_resource_start(pdev, 0);

			if (mb->iobase[0] == 0)
			{
				printk("SUNIX Error: SUNIX Matrix Board (bus:%d device:%d), in configuartion space, Bar 0 isn't valid !\n\n",mb->bus_number,mb->dev_number);
				status = STATUS_PCICONFSPACE_BAR0_ERR;
				return status;
			}

			for (j = 0; j < mb->ports; j++)
			{
				mb->iobase[j] = mb->iobase[0] + mb->pb_info.port[j].offset;
				mb->ChipFlag[j] = MATRIX_HW_ID_TYPE_I;
			}

			mb->vector_mask = 0;

			for (j = 0; j < mb->ports; j++)
			{
				mb->vector_mask  |= (1 << j);
			}

			mb->vector = pci_resource_start(pdev, 1);

			if (mb->vector == 0)
			{
				printk("SUNIX Error: SUNIX Matrix Board (bus:%d device:%d), in configuartion space, interrupt vector isn't valid !\n\n",mb->bus_number,mb->dev_number);
				status = STATUS_PCICONFSPACE_VETOR_ERR;
				return status;
			}

			mb->irq = mb->pdev->irq;

			if (mb->irq <= 0)
			{
				printk("SUNIX Error: SUNIX Matrix Board (bus:%d device:%d), in configuartion space, irq isn't valid !\n\n",mb->bus_number,mb->dev_number);
				status = STATUS_PCICONFSPACE_IRQ_ERR;
				return status;
			}

			request_region(mb->iobase[0], MATRIX_TYPE_I_BAR0_LENGTH, "sunix_matrix");
			request_region(mb->vector, MATRIX_TYPE_I_BAR1_LENGTH, "sunix_matrix");

			matrix_drvconf.board_count++;
		}
	}

	return status;
}


static int matrix_pci_board_probe(void)
{
	struct MATRIX_BOARD *mb;
	struct pci_dev *pdev = NULL;
	struct pci_dev *pdev_array[4] = {NULL, NULL, NULL, NULL};

	int matrix_pci_board_id_cnt;
	int tablecnt = 0;
	int boardcnt = 0;
	int i;

	unsigned short int subdeviceID = 0;
	int status = STATUS_SUCCESS;

	memset(matrix_pci_board_table, 0, MATRIX_PCI_BOARDS_MAX * sizeof(struct MATRIX_BOARD));

	for (i = 0; i < MATRIX_PCI_BOARDS_MAX; i++)
	{
		matrix_pci_board_table[i].board_enum = -1;
		matrix_pci_board_table[i].board_number = -1;
	}

	matrix_pci_board_id_cnt = (sizeof(matrix_pci_board_id) / sizeof(matrix_pci_board_id[0])) - 1;

	i=1;

	while (tablecnt < matrix_pci_board_id_cnt)
	{
		pdev = pci_find_device(VENID_MATRIX, DEVID_MATRIX_TYPE_I, pdev);

		if (pdev == NULL)
		{
			tablecnt++;
			continue;
		}

		if ((tablecnt > 0) && ((pdev == pdev_array[0]) || (pdev == pdev_array[1]) || (pdev == pdev_array[2]) || (pdev == pdev_array[3])))
		{
			continue;
		}

		pci_read_config_word(pdev, 0x2e, &subdeviceID);

		if (subdeviceID == 0)
		{
			printk("SUNIX Error: SUNIX Matrix Board (bus:%d device:%d), in configuration space,\n",pdev->bus->number,PCI_SLOT(pdev->devfn));
			printk("             subdevice id isn't vaild.\n\n");
			status = STATUS_SUBDEVICEID_ERR;
			return status;
		}

		if (subdeviceID != matrix_pci_board_id[tablecnt].subdevice)
		{
			continue;
		}

		if (pdev == NULL)
		{
			printk("SUNIX Error: PCI device object is an NULL pointer !\n\n");
			status = STATUS_PDEV_EQL_NULL;
			return status;
		}
		else
		{
			status = pci_enable_device(pdev);

			if (status != 0)
			{
				printk("SUNIX Error: SUNIX Matrix Board Enable Fail !\n\n");
				matrix_drvconf.flags |= DRIVER_PCI_DEV_ENABLED_ERR;
				status = STATUS_PCIDEV_ENABLE_FAIL;
				return status;
			}
		}

		boardcnt++;

		if (boardcnt > MATRIX_PCI_BOARDS_MAX)
		{
			printk("\n");
 			printk("SUNIX Error: SUNIX Matrix Series Driver Module Support Four Boards In Maximum !\n\n");	
			status = STATUS_BOARD_EXCEED;
 			return status;
 		}

		mb = &matrix_pci_board_table[boardcnt-1];

		pdev_array[boardcnt-1] = pdev;
		mb->pdev = pdev;
		mb->bus_number = pdev->bus->number;
 		mb->dev_number = PCI_SLOT(pdev->devfn);

		mb->board_enum = (int)matrix_pci_board_id[tablecnt].driver_data;
		mb->pb_info = matrix_pci_board_conf[matrix_pci_board_id[tablecnt].driver_data];
		mb->board_number = 	boardcnt-1;

		printk("\n");
		printk("SUNIX Info : Found SUNIX Matrix Series %d ports %s Board,\n",mb->pb_info.numberOfPorts,mb->pb_info.functionName);
		printk("             bus number:%d, device number:%d\n\n",mb->bus_number,mb->dev_number);
	}


	if (boardcnt == 0)
	{
		printk("SUNIX Info : No SUNIX Matrix Series Board Found !\n\n");
		status = STATUS_NO_BOARD_FOUND;
		return status;
	}

	return status;
}


static int matrix_pci_board_release(void)
{
	struct MATRIX_BOARD *mb = NULL;
	struct MATRIX_UART_PORT *up = NULL;
	int i;
	int status = STATUS_SUCCESS;

	for (i = 0; i < MATRIX_PCI_BOARDS_MAX; i++)
	{
		mb = &matrix_pci_board_table[i];

		if (mb->board_enum > 0)
		{
			{
				up = &matrix_port_table[mb->port_index];
				free_irq(mb->irq, up);

				mb->flags &= ~PCIBOARD_IRQ_REGISTER_DONE;
			}

			{
				release_region(mb->iobase[0], MATRIX_TYPE_I_BAR0_LENGTH);
				release_region(mb->vector, MATRIX_TYPE_I_BAR1_LENGTH);
				mb->flags &= ~PCIBOARD_IO_REQUEST_DONE;
			}

			{
				pci_disable_device(mb->pdev);
				mb->flags &= ~PCIBOARD_ENABLED_DONE;
			}

		}
	}

	return status;
}


int init_module(void)
{
	int status = STATUS_SUCCESS;
	int fail_status = STATUS_SUCCESS;

	printk("\n\n");
	printk("=================  SUNIX Matrix Series Driver Module Install  ==================\n");
	printk("\n");
	printk("SUNIX Info : Loading SUNIX Matrix Series Board Driver Module\n");
	printk("                                                       -- Version: %s\n\n",DRIVER_VERSION);

	memset(&matrix_drvconf, 0, sizeof(struct MATRIX_DRIVER_CONF));

	spin_lock_init(&matrix_drvconf.lock);

	status = matrix_register_tty_driver();
	if (status != STATUS_SUCCESS)
	{
		goto init_fail;
	}

	status = matrix_pci_board_probe();
	if (status != STATUS_SUCCESS)
	{
		goto init_fail;
	}

	status = matrix_get_pci_board_conf();
	if (status != STATUS_SUCCESS)
	{
		goto init_fail;
	}

	status = matrix_board_init();
	if (status != STATUS_SUCCESS)
	{
		goto init_fail;
	}

	printk("================================================================================\n");
	return status;


init_fail:

	{
		fail_status = matrix_unregister_tty_driver();
	}

	{
		fail_status = matrix_pci_board_release();
	}

	printk("SUNIX Error: Couldn't Loading SUNIX Matrix Series Board Driver Module correctly\n");
	printk("             , please reboot system and try again, if error still happen\n");
	printk("             , contact SUNIX support.\n\n");
	printk("================================================================================\n");
	return status;
}


void cleanup_module(void)
{
	int status = STATUS_SUCCESS;

	printk("\n\n");
	printk("================  SUNIX Matrix Series Driver Module Uninstall  =================\n");
	printk("\n");

	status = matrix_unregister_tty_driver();

	status = matrix_pci_board_release();

	printk("SUNIX Info : Unload SUNIX Matrix Series Board Driver Module Done.\n");
	printk("================================================================================\n");
}
